/*
 * (SLIK) SimpLIstic sKin functions
 * (C) 2001 John Ellis
 *
 * Author: John Ellis
 *
 * This software is released under the GNU General Public License (GNU GPL).
 * Please read the included file COPYING for more information.
 * This software comes with no warranty of any kind, use at your own risk!
 */

#include "ui2_includes.h"
#include "ui2_typedefs.h"

#include "ui2_text.h"
#include "ui2_text_edit.h"

#include "ui2_display.h"
#include "ui2_editor.h"
#include "ui2_widget.h"
#include "ui_fileops.h"
#include "ui_pixbuf_ops.h"
#include "ui_tabcomp.h"
#include "ui_utildlg.h"


typedef struct _TextDetail TextDetail;
struct _TextDetail
{
	TextData *text;
	GtkWidget *spin_length;
	GtkWidget *extended_button;

	WidgetData *wd;
	UIData *ui;
};

typedef struct _TextListData TextListData;
struct _TextListData
{
	gchar *key;
	gchar *text_id;
	gchar *data;

	gchar *image;
	gint extended;
	gint length;
};

typedef struct _TextPage TextPage;
struct _TextPage
{
	GtkWidget *key_entry;
	GtkWidget *text_id_entry;
	GtkWidget *data_entry;

	GtkWidget *image_entry;
	GtkWidget *extended_button;
	GtkWidget *length_spin;

	GtkWidget *clist;
	GtkWidget *pixmap;

	EditData *ed;
};

static GdkPixbuf *text_get_pixbuf(gpointer data)
{
	TextData *text = data;
	return text->overlay;
}

static void text_edit_write(FILE *f, WidgetData *wd, SkinData *skin, const gchar *dir)
{
	TextData *text = wd->widget;
	gchar *image;

	image = ui_edit_copy_unique_file(ui_widget_get_data(wd, "image"),
					 dir, text->overlay, "text_", wd->key);

	if (image) ui_edit_widget_set_path_key(wd, "image", dir, image);

	ui_edit_write_section(f, "text", wd->key);

	ui_edit_write_key_char(f, "image", image);
	ui_edit_write_key_int(f, "x", text->x);
	ui_edit_write_key_int(f, "y", text->y);
	ui_edit_write_key_int(f, "length", text->length);

	ui_edit_write_key_bool(f, "extended", text->extended);

	g_free(image);
}

static TextListData *text_edit_find(GList *list, const gchar *image)
{
	GList *work;
	work = list;
	while(work)
		{
		TextListData *td = work->data;
		if (strcmp(image, td->image) == 0) return td;
		work = work->next;
		}
	return NULL;
}

static gpointer text_edit_read(UIData *ui, WidgetData *wd, GList *list)
{
	TextData *text = wd->widget;
	TextListData *td;
	const gchar *image;

	image = ui_widget_get_data(wd, "image");

	if (!image || text_edit_find(list, image)) return NULL;

	td = g_new0(TextListData, 1);
	td->image = g_strdup(image);
	td->extended = text->extended;
	td->length = text->length;

	td->key = g_strdup(wd->key);
	td->data = g_strdup(ui_widget_get_data(wd, "data"));
	td->text_id = g_strdup(wd->text_id);

	return td;
}

static void text_edit_free(gpointer data)
{
	TextListData *td = data;

	g_free(td->key);
	g_free(td->text_id);
	g_free(td->data);
	g_free(td->image);
	g_free(td);
}

static void text_edit_props_length_cb(GtkAdjustment *adjustment, gpointer data)
{
	TextDetail *td = data;
	gint new_length;
	gint old_w, w;

	new_length = ui_edit_spin_get(td->spin_length);

	if (new_length * td->text->char_width > td->ui->skin->width)
		{
		new_length = td->ui->skin->width / td->text->char_width;
		}

	if (new_length == td->text->length) return;

	old_w = td->text->width;
	w = td->text->char_width * new_length;

	if (new_length > td->text->length &&
	    td->text->x + w > td->ui->skin->width)
		{
		ui_widget_set_coord(td->ui, td->wd, td->ui->skin->width - w, td->text->y, TRUE);
		}

	ui_edit_widget_draw_highlight(td->ui, td->wd, FALSE);
	pixbuf_copy_area(td->ui->skin->overlay, td->text->x, td->text->y,
			 td->ui->skin->pixbuf, td->text->x, td->text->y, old_w, td->text->height, FALSE);

	td->text->length = new_length;
	td->text->width = w;
	gdk_pixbuf_unref(td->text->pixbuf);
	td->text->pixbuf = gdk_pixbuf_new(GDK_COLORSPACE_RGB, FALSE, 8, w, td->text->height);
	
	ui_widget_sync_back(td->ui, td->wd);
	ui_widget_draw(td->ui, td->wd, TRUE, TRUE);
	ui_edit_widget_draw_highlight(td->ui, td->wd, TRUE);
	ui_display_render_area(td->ui, td->text->x, td->text->y, old_w, td->text->height);
}

static gpointer text_edit_props(UIData *ui, WidgetData *wd, GtkWidget *vbox, gpointer detail)
{
	TextData *text = wd->widget;
	TextDetail *td = detail;

	if (!td)
		{
		GtkObject *adj;

		td = g_new0(TextDetail, 1);

		td->extended_button = ui_edit_toggle_new(vbox, _("Extended"));
		gtk_widget_set_sensitive(td->extended_button, FALSE);

		td->spin_length = ui_edit_spin_new(vbox, _("length:"), 1, 200, &adj);
		gtk_signal_connect(adj, "value_changed", text_edit_props_length_cb, td);
		}

	td->text = text;
	td->wd = wd;
	td->ui = ui;

	ui_edit_toggle_set(td->extended_button, text->extended);

	ui_edit_spin_set_blocking(td->spin_length, text->length, td);

	return td;
}

static void text_edit_page_add_cb(GtkWidget *widget, gpointer data)
{
	TextPage *pd = data;
	TextData *text;
	const gchar *key;
	const gchar *text_id;
	const gchar *image;

	key = ui_edit_entry_get(pd->key_entry);
	text_id = ui_edit_entry_get(pd->text_id_entry);
	image = ui_edit_entry_get(pd->image_entry);

	if (!key || !image || !isfile(image))
		{
		warning_dialog(_("Text error"), _("Text must contain a key and valid image."));
		return;
		}

	text = text_new_from_file(image, ui_edit_spin_get(pd->length_spin),
				  ui_edit_toggle_get(pd->extended_button) , 0, 0);
	if (text)
		{
		WidgetData *wd;
		wd = text_register(pd->ed->ui->skin, text, key, text_id);
		ui_edit_widget_add_finish(pd->ed, wd, image, ui_edit_entry_get(pd->data_entry));
		}

	tab_completion_append_to_history(pd->image_entry, image);
}

static void text_edit_page_sync(TextPage *pd, TextListData *td)
{
	if (!td) return;

	ui_edit_entry_set(pd->key_entry, td->key);
	ui_edit_entry_set(pd->text_id_entry, td->text_id);
	ui_edit_entry_set(pd->data_entry, td->data);

	ui_edit_entry_set(pd->image_entry, td->image);

	ui_edit_toggle_set(pd->extended_button, td->extended);
	ui_edit_spin_set(pd->length_spin, td->length);
}

static void text_edit_page_clist_cb(GtkWidget *clist, gint row, gint col, GdkEvent *event, gpointer data)
{
	TextPage *pd = data;
	TextListData *td;

	td = gtk_clist_get_row_data(GTK_CLIST(clist), row);
	text_edit_page_sync(pd, td);
}

static void text_edit_page_destroy_cb(GtkWidget *widget, gpointer data)
{
	TextPage *pd = data;

	g_free(pd);
}

static GtkWidget *text_edit_page_new(EditData *ed)
{
	GtkWidget *hbox;
	GtkWidget *vbox;
	GtkWidget *button;
	TextPage *pd;
	gchar *titles[] = { _("Image"), _("Key"), _("Extended"), NULL };

	pd = g_new0(TextPage, 1);
	pd->ed = ed;

	hbox = gtk_hbox_new(FALSE, 5);
	gtk_container_set_border_width(GTK_CONTAINER(hbox), 5);
	gtk_object_set_data(GTK_OBJECT(hbox), "page", pd);
	gtk_signal_connect(GTK_OBJECT(hbox), "destroy",
			   GTK_SIGNAL_FUNC(text_edit_page_destroy_cb), pd);

	vbox = gtk_vbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(hbox), vbox, FALSE, FALSE, 0);
	gtk_widget_show(vbox);

	pd->key_entry = ui_edit_key_entry_new(vbox, ed->ui, text_type_id());
	pd->data_entry = ui_edit_entry_new(vbox, _("Data:"));
	pd->text_id_entry = ui_edit_entry_new(vbox, _("Text id:"));

	pd->extended_button = ui_edit_toggle_new(vbox, _("Extended"));
	pd->length_spin = ui_edit_spin_new(vbox, _("Length:"), 1, 200, NULL);

	pd->image_entry = ui_edit_path_entry_new(vbox, _("Image:"), "SLIK_text_image");
	pd->pixmap = ui_edit_pixmap_new(vbox);
	ui_edit_path_entry_connect_pixmap(pd->image_entry, pd->pixmap);

	button = gtk_button_new_with_label(_("Add"));
	gtk_box_pack_start(GTK_BOX(vbox), button, FALSE, FALSE, 0);
	gtk_signal_connect(GTK_OBJECT(button), "clicked", text_edit_page_add_cb, pd);
	gtk_widget_show(button);

	pd->clist = ui_edit_clist_new(hbox, titles, 3);
	gtk_signal_connect(GTK_OBJECT(pd->clist), "select_row", text_edit_page_clist_cb, pd);

	gtk_widget_show(hbox);

	return hbox;
}

static void text_edit_page_add(GtkWidget *widget, gpointer data)
{
	TextListData *td = data;
	TextPage *pd;
	gchar *buf[4];

	pd = gtk_object_get_data(GTK_OBJECT(widget), "page");

	buf[0] = "";
	buf[1] = td->key;
	buf[2] = ui_edit_bool_to_text(td->extended);
	buf[3] = NULL;

	ui_edit_clist_append(pd->clist, buf, td->image, td);
}

void text_type_init_edit(WidgetObjectData *od)
{
	od->func_get_pixbuf = text_get_pixbuf;

	od->func_edit_write = text_edit_write;

	od->func_edit_read = text_edit_read;
	od->func_edit_free = text_edit_free;

	od->func_edit_props = text_edit_props;

	od->func_edit_page_new = text_edit_page_new;
	od->func_edit_page_add = text_edit_page_add;
}


